"""Utilities for generating predictions with lexical predictions."""

from __future__ import annotations

import itertools as itt
import logging
import typing
from collections import defaultdict
from collections.abc import Callable, Iterable, Mapping
from pathlib import Path
from typing import TYPE_CHECKING, TypeAlias, cast

import click
import curies
import ssslm
import sssom_pydantic
from bioregistry import NormalizedNamedReference
from curies.vocabulary import lexical_matching_process
from more_click import verbose_option
from sssom_pydantic import MappingTool, SemanticMapping
from tqdm.auto import tqdm

from .embedding import predict_embedding_mappings
from .utils import TOOL_NAME, resolve_mapping_tool, resolve_predicate
from ..constants import PredictionMethod, RecognitionMethod

if TYPE_CHECKING:
    import gilda
    import networkx as nx

__all__ = [
    "TOOL_NAME",
    "append_lexical_predictions",
    "append_predictions",
    "filter_custom",
    "filter_existing_xrefs",
    "get_predictions",
    "lexical_prediction_cli",
    "predict_embedding_mappings",
    "predict_lexical_mappings",
]

logger = logging.getLogger(__name__)

#: A filter 3-dictionary of source prefix to target prefix to source identifier to target identifier
NestedMappingDict: TypeAlias = Mapping[str, Mapping[str, Mapping[str, str]]]

#: A callable that gets matches
MatchCallable: TypeAlias = Callable[[str], list[ssslm.Match]]


def get_predictions(
    prefix: str,
    target_prefixes: str | Iterable[str],
    *,
    relation: str | None | curies.NamableReference = None,
    identifiers_are_names: bool = False,
    method: PredictionMethod | None = None,
    cutoff: float | None = None,
    batch_size: int | None = None,
    custom_filter_function: Callable[[SemanticMapping], bool] | None = None,
    progress: bool = True,
    filter_mutual_mappings: bool = False,
    mapping_tool: str | MappingTool | None = None,
    force: bool = False,
    cache: bool = True,
    force_process: bool = False,
    all_by_all: bool = False,
) -> list[SemanticMapping]:
    """Add lexical matching-based predictions to the Biomappings predictions.tsv file.

    :param prefix: The source prefix
    :param target_prefixes: The target prefix or prefixes
    :param mapping_tool: The name of the mapping tool. Defaults to :data:``
    :param relation: The relationship. Defaults to ``skos:exactMatch``.
    :param identifiers_are_names: The source prefix's identifiers should be considered
        as names
    :param method: The lexical predication method to use
    :param cutoff: an optional minimum prediction confidence cutoff
    :param batch_size: The batch size for embeddings
    :param custom_filter_function: A custom function that decides if semantic mappings
        should be kept, applied after all other logic.
    :param progress: Should progress be shown?
    :param filter_mutual_mappings: Should mappings between entities in the given
        namespaces be filtered out?
    :param force: Should the ontologies be re-downloaded/processed?
    :param force_process: Should the ontologies be re-processed? Subsumed by ``force``
    :param cache: Should the results of processing the resources be cached?
    :param all_by_all: Enable all-by-all prediction mode, which doesn't just check
        between the given prefix and target_prefixes (1-n) but does all against all

    :returns: A list of predicted semantic mappings
    """
    if isinstance(target_prefixes, str):
        targets = [target_prefixes]
    else:
        targets = list(target_prefixes)

    mapping_tool = resolve_mapping_tool(mapping_tool)

    if method is None or method in typing.get_args(RecognitionMethod):
        import pyobo

        if all_by_all:
            grounder = pyobo.get_grounder(
                [prefix, *targets],
                raise_on_missing=False,
                force=force,
                force_process=force_process,
                cache=cache,
            )
            predictions = _predict_lexical_mappings_all_by_all(
                grounder,
                predicate=relation,
                mapping_tool=mapping_tool,
                method=cast(RecognitionMethod | None, method),
            )
        else:
            # by default, PyOBO wraps a gilda grounder, but
            # can be configured to use other NER/NEN systems
            grounder = pyobo.get_grounder(
                targets,
                raise_on_missing=False,
                force=force,
                force_process=force_process,
                cache=cache,
            )
            predictions = predict_lexical_mappings(
                prefix,
                predicate=relation,
                grounder=grounder,
                mapping_tool=mapping_tool,
                identifiers_are_names=identifiers_are_names,
                method=cast(RecognitionMethod | None, method),
            )
    elif method == "embedding":
        if all_by_all:
            raise NotImplementedError(
                "all-by-all prediction not implemented for embedding workflow"
            )
        predictions = predict_embedding_mappings(
            prefix,
            target_prefixes,
            mapping_tool=mapping_tool,
            relation=relation,
            cutoff=cutoff,
            batch_size=batch_size,
            progress=progress,
        )
    else:
        raise ValueError(f"invalid lexical prediction method: {method}")

    if filter_mutual_mappings:
        mutual_mapping_filter = _get_mutual_mapping_filter(prefix, target_prefixes)
        predictions = filter_custom(predictions, mutual_mapping_filter)

    predictions = filter_existing_xrefs(predictions, [prefix, *targets])

    if custom_filter_function:
        predictions = list(filter(custom_filter_function, predictions))

    predictions = sorted(predictions)
    return predictions


def _get_get_matches(method: RecognitionMethod | None, grounder: ssslm.Grounder) -> MatchCallable:
    if method is None or method == "grounding":
        return grounder.get_matches
    elif method == "ner":

        def _get_matches(s: str) -> list[ssslm.Match]:
            return [a.match for a in grounder.annotate(s)]

        return _get_matches

    else:
        raise ValueError(f"invalid lexical method: {method}")


def _predict_lexical_mappings_all_by_all(
    grounder: ssslm.Grounder,
    *,
    predicate: str | curies.Reference | None = None,
    method: RecognitionMethod | None = None,
    mapping_tool: str | MappingTool | None = None,
) -> Iterable[SemanticMapping]:
    """Iterate over predictions."""
    predicate = resolve_predicate(predicate)
    mapping_tool = resolve_mapping_tool(mapping_tool)
    if method not in {None, "grounding"}:
        raise NotImplementedError(f"all-by-all requires grounding method, {method} not allowed")
    if not isinstance(grounder, ssslm.GildaGrounder):
        raise NotImplementedError(f"all-by-all requires gilda grounder. got: {type(grounder)}")
    yield from _all_by_all_gilda(grounder._grounder, predicate, mapping_tool)


def _all_by_all_gilda(
    grounder: gilda.Grounder, predicate: curies.Reference, mapping_tool: MappingTool | None = None
) -> Iterable[SemanticMapping]:
    from gilda.scorer import generate_match
    from gilda.scorer import score as get_score

    for values in grounder.entries.values():
        for s, o in itt.combinations(values, 2):
            if s.db == o.db:
                continue
            match = generate_match(s.text, o.text)
            score = get_score(match, o)  # FIXME not symmetric
            yield SemanticMapping(
                subject=NormalizedNamedReference(prefix=s.db, identifier=s.id, name=s.entry_name),
                predicate=predicate,
                object=NormalizedNamedReference(prefix=o.db, identifier=o.id, name=o.entry_name),
                justification=lexical_matching_process,
                confidence=round(score, 3),
                mapping_tool=mapping_tool,
            )


def predict_lexical_mappings(
    prefix: str,
    *,
    predicate: str | curies.Reference | None = None,
    grounder: ssslm.Grounder,
    identifiers_are_names: bool = False,
    strict: bool = False,
    method: RecognitionMethod | None = None,
    mapping_tool: str | MappingTool | None = None,
) -> Iterable[SemanticMapping]:
    """Iterate over prediction tuples for a given prefix."""
    import pyobo

    id_name_mapping = pyobo.get_id_name_mapping(prefix, strict=strict)
    it = tqdm(
        id_name_mapping.items(), desc=f"[{prefix}] lexical tuples", unit_scale=True, unit="name"
    )

    predicate = resolve_predicate(predicate)
    get_matches = _get_get_matches(method, grounder)
    mapping_tool = resolve_mapping_tool(mapping_tool)

    name_prediction_count = 0
    for identifier, name in it:
        for scored_match in get_matches(name):
            name_prediction_count += 1
            yield SemanticMapping(
                subject=NormalizedNamedReference(prefix=prefix, identifier=identifier, name=name),
                predicate=predicate,
                object=scored_match.reference,
                justification=lexical_matching_process,
                confidence=round(scored_match.score, 3),
                mapping_tool=mapping_tool,
            )

    tqdm.write(f"[{prefix}] generated {name_prediction_count:,} predictions from names")

    if identifiers_are_names:
        it = tqdm(
            pyobo.get_ids(prefix), desc=f"[{prefix}] lexical tuples", unit_scale=True, unit="id"
        )
        identifier_prediction_count = 0
        for identifier in it:
            for scored_match in get_matches(identifier):
                name_prediction_count += 1
                yield SemanticMapping(
                    subject=NormalizedNamedReference(
                        prefix=prefix, identifier=identifier, name=identifier
                    ),
                    predicate=predicate,
                    object=scored_match.reference,
                    justification=lexical_matching_process,
                    confidence=round(scored_match.score, 3),
                    mapping_tool=mapping_tool,
                )
        tqdm.write(
            f"[{prefix}] generated {identifier_prediction_count:,} predictions from identifiers"
        )


def filter_custom(
    mappings: Iterable[SemanticMapping],
    custom_filter: NestedMappingDict,
) -> Iterable[SemanticMapping]:
    """Filter out custom mappings."""
    counter = 0
    for mapping in mappings:
        if (
            custom_filter.get(mapping.subject.prefix, {})
            .get(mapping.object.prefix, {})
            .get(mapping.subject.identifier)
        ):
            counter += 1
            continue
        yield mapping
    logger.info("filtered out %d custom mapped matches", counter)


def filter_existing_xrefs(
    mappings: Iterable[SemanticMapping], prefixes: Iterable[str]
) -> Iterable[SemanticMapping]:
    """Filter predictions that match xrefs already loaded through PyOBO.

    :param mappings: Semantic mappings to filter
    :param prefixes: Prefixes for resources to check for existing mappings

    :yields: Filtered semantic mappings
    """
    entity_to_mapped_prefixes = _get_entity_to_mapped_prefixes(prefixes)

    n_predictions = 0
    for mapping in tqdm(mappings, desc="filtering predictions", leave=False):
        if (
            mapping.subject in entity_to_mapped_prefixes
            and mapping.object.prefix in entity_to_mapped_prefixes[mapping.subject]
        ) or (
            mapping.object in entity_to_mapped_prefixes
            and mapping.subject.prefix in entity_to_mapped_prefixes[mapping.object]
        ):
            n_predictions += 1
            continue
        yield mapping

    if n_predictions:
        tqdm.write(f"filtered out {n_predictions:,} pre-mapped matches")


def _get_entity_to_mapped_prefixes(prefixes: Iterable[str]) -> dict[curies.Reference, set[str]]:
    import pyobo

    entity_to_mapped_prefixes: defaultdict[curies.Reference, set[str]] = defaultdict(set)
    for prefix in prefixes:
        for mapping in pyobo.get_semantic_mappings(prefix):
            entity_to_mapped_prefixes[mapping.subject].add(mapping.object.prefix)
            entity_to_mapped_prefixes[mapping.object].add(mapping.subject.prefix)
    return dict(entity_to_mapped_prefixes)


def _get_mutual_mapping_filter(prefix: str, targets: str | Iterable[str]) -> NestedMappingDict:
    """Get a custom filter dictionary.

    This is induced over the mutual mapping graph with all target prefixes.

    :param prefix: The source prefix
    :param targets: All potential target prefixes

    :returns: A filter 3-dictionary of source prefix to target prefix to source
        identifier to target identifier
    """
    try:
        import networkx as nx
    except ImportError as e:
        raise ImportError(
            "NetworkX is required for mapping filtering, install with pip install networkx"
        ) from e

    if isinstance(targets, str):
        targets = [targets]
    graph = _mutual_mapping_graph([prefix, *targets])
    rv: defaultdict[str, dict[str, str]] = defaultdict(dict)
    for node in graph:
        if node.prefix != prefix:
            continue
        for xref_prefix, xref_identifier in nx.single_source_shortest_path(graph, node):
            rv[xref_prefix][node.identifier] = xref_identifier
    return {prefix: dict(rv)}


def _mutual_mapping_graph(prefixes: Iterable[str]) -> nx.Graph:
    """Get the undirected mapping graph between the given prefixes.

    :param prefixes: A list of prefixes to use with :func:`pyobo.get_filtered_xrefs` to
        get xrefs.

    :returns: The undirected mapping graph containing mappings between entries in the
        given namespaces.
    """
    import networkx as nx
    import pyobo

    prefixes = set(prefixes)
    graph = nx.Graph()
    for prefix in sorted(prefixes):
        for mapping in pyobo.get_semantic_mappings(prefix):
            if mapping.object.prefix not in prefixes:
                continue
            graph.add_edge(mapping.subject, mapping.object)
    return graph


def _upgrade_set(values: Iterable[str] | None = None) -> set[str]:
    return set() if values is None else set(values)


def append_predictions(
    new_mappings: Iterable[SemanticMapping],
    *,
    path: Path,
    curated_paths: list[Path] | None = None,
    converter: curies.Converter | None = None,
) -> None:
    """Append new lines to the predictions table."""
    mappings, converter, metadata = sssom_pydantic.read(path, converter=converter)

    prefixes: set[str] = set()
    for mapping in new_mappings:
        prefixes.update(mapping.get_prefixes())
        mappings.append(mapping)

    prefixes_to_add = {prefix for prefix in prefixes if not converter.standardize_prefix(prefix)}
    if prefixes_to_add:
        try:
            import bioregistry
        except ImportError as e:
            raise ImportError("amending prefixes requires the bioregistry to be installed") from e

        for prefix in prefixes_to_add:
            resource = bioregistry.get_resource(prefix)
            if resource is None:
                raise ValueError(
                    f"can not automatically extend because {prefix} is not registered in the "
                    f"bioregistry.\n\nSolution: add a CURIE prefix-URI prefix mapping manually "
                    f"to the file {path}"
                )

            uri_prefix = resource.get_rdf_uri_prefix() or resource.get_uri_prefix()
            if not uri_prefix:
                raise ValueError(
                    f"can not automatically extend because {prefix} does not have a valid URI "
                    f"prefix in the bioregistry.\n\nSolution: add a CURIE prefix-URI prefix "
                    f"mapping manually to the file {path}"
                )
            converter.add_prefix(
                resource.prefix,
                uri_prefix,
            )

    if curated_paths is not None:
        exclude_mappings = itt.chain.from_iterable(
            sssom_pydantic.read(path)[0] for path in curated_paths
        )
    else:
        exclude_mappings = None

    sssom_pydantic.write(
        mappings,
        path,
        metadata=metadata,
        converter=converter,
        drop_duplicates=True,
        sort=True,
        exclude_mappings=exclude_mappings,
    )


def append_lexical_predictions(
    prefix: str,
    target_prefixes: str | Iterable[str],
    *,
    relation: str | None | curies.NamableReference = None,
    identifiers_are_names: bool = False,
    path: Path,
    method: PredictionMethod | None = None,
    cutoff: float | None = None,
    batch_size: int | None = None,
    custom_filter_function: Callable[[SemanticMapping], bool] | None = None,
    progress: bool = True,
    filter_mutual_mappings: bool = False,
    curated_paths: list[Path] | None = None,
    mapping_tool: str | MappingTool | None = None,
    force: bool = False,
    force_process: bool = False,
    cache: bool = True,
    converter: curies.Converter | None = None,
    all_by_all: bool = False,
) -> None:
    """Add lexical matching-based predictions to the Biomappings predictions.tsv file.

    :param prefix: The source prefix
    :param target_prefixes: The target prefix or prefixes
    :param relation: The relationship. Defaults to ``skos:exactMatch``.
    :param identifiers_are_names: The source prefix's identifiers should be considered
        as names
    :param path: A custom path to predictions TSV file
    :param method: The lexical predication method to use
    :param cutoff: an optional minimum prediction confidence cutoff
    :param batch_size: The batch size for embeddings
    :param custom_filter_function: A custom function that decides if semantic mappings
        should be kept, applied after all other logic.
    :param progress: Should progress be shown?
    :param filter_mutual_mappings: Should mappings between entities in the given
        namespaces be filtered out?
    :param mapping_tool: The name of the mapping tool
    :param curated_paths: The paths to curated documents that are used to remove zombie
        mappings (i.e., predictions that were already curated)
    """
    predictions = get_predictions(
        prefix,
        target_prefixes,
        mapping_tool=mapping_tool,
        relation=relation,
        identifiers_are_names=identifiers_are_names,
        method=method,
        cutoff=cutoff,
        batch_size=batch_size,
        custom_filter_function=custom_filter_function,
        progress=progress,
        filter_mutual_mappings=filter_mutual_mappings,
        force=force,
        force_process=force_process,
        cache=cache,
        all_by_all=all_by_all,
    )
    tqdm.write(f"[{prefix}] generated {len(predictions):,} predictions")

    # since the function that constructs the predictions already
    # pre-standardizes, we don't have to worry about standardizing again
    append_predictions(predictions, path=path, curated_paths=curated_paths, converter=converter)


def lexical_prediction_cli(
    prefix: str,
    target: str | list[str],
    *,
    path: Path,
    curated_paths: list[Path] | None = None,
    filter_mutual_mappings: bool = False,
    identifiers_are_names: bool = False,
    predicate: str | None | curies.NamableReference = None,
    method: PredictionMethod | None = None,
    cutoff: float | None = None,
    custom_filter_function: Callable[[SemanticMapping], bool] | None = None,
    mapping_tool: str | MappingTool | None = None,
) -> None:
    """Construct a CLI and run it."""
    tt = target if isinstance(target, str) else ", ".join(target)

    @click.command(help=f"Generate mappings from {prefix} to {tt}")
    @click.option("--force", is_flag=True)
    @verbose_option
    def main(force: bool) -> None:
        """Generate mappings."""
        append_lexical_predictions(
            prefix,
            target,
            path=path,
            curated_paths=curated_paths,
            filter_mutual_mappings=filter_mutual_mappings,
            identifiers_are_names=identifiers_are_names,
            relation=predicate,
            method=method,
            cutoff=cutoff,
            custom_filter_function=custom_filter_function,
            mapping_tool=mapping_tool,
            force=force,
        )

    main()



"""Utilities for lexical prediction workflows."""

from __future__ import annotations

import curies
from bioregistry import NormalizedReference
from curies.vocabulary import exact_match
from sssom_pydantic import MappingTool

__all__ = [
    "TOOL_NAME",
    "resolve_mapping_tool",
    "resolve_predicate",
]

#: The name of the lexical mapping tool
TOOL_NAME = "sssom-curator"


def resolve_mapping_tool(mapping_tool: str | MappingTool | None) -> MappingTool:
    """Resolve the mapping tool."""
    if mapping_tool is None:
        return MappingTool(name=TOOL_NAME, version=None)
    if isinstance(mapping_tool, str):
        return MappingTool(name=mapping_tool, version=None)
    return mapping_tool


def resolve_predicate(predicate: str | curies.Reference | None = None) -> NormalizedReference:
    """Ensure a predicate is available."""
    if predicate is None:
        predicate = exact_match
    elif isinstance(predicate, str):
        predicate = NormalizedReference.from_curie(predicate)

    # throw away name so we don't make a label column
    predicate = NormalizedReference(prefix=predicate.prefix, identifier=predicate.identifier)
    return predicate



"""Initialize repositories."""

from __future__ import annotations

import os
import stat
from pathlib import Path
from typing import TYPE_CHECKING

import click

if TYPE_CHECKING:
    import jinja2
    from sssom_pydantic import MappingSet

    from ..repository import Repository

__all__ = [
    "initialize_folder",
]

HERE = Path(__file__).parent.resolve()
SCRIPT_NAME = "main.py"
README_NAME = "README.md"
CC0_CURIE = "spdx:CC0-1.0"
DATA_DIR_NAME = "data"
SKIPS = {
    "mapping_set": {
        "extension_definitions",
        "creator_label",
        "publication_date",
        "sssom_version",
        "issue_tracker",
        "other",
    }
}


def normalize_name(name: str) -> str:
    """Normalize a name by replacing spaces and underscores with dashes."""
    return name.replace(" ", "-").replace("_", "-").lower()


def initialize_folder(  # noqa:C901
    directory: str | Path,
    *,
    positive_mappings_filename: str | None = None,
    unsure_mappings_filename: str | None = None,
    predicted_mappings_filename: str | None = None,
    negative_mappings_filename: str | None = None,
    repository_filename: str | None = None,
    mapping_set: MappingSet | None = None,
    purl_base: str | None = None,
    script_filename: str = SCRIPT_NAME,
    readme_filename: str = README_NAME,
    add_license: bool = True,
    mapping_set_id: str | None = None,
) -> Repository:
    """Create a curation repository in a folder.

    :param directory: The directory where to create it.

    Creates the following:

    1. Four curation files, each loaded up with Bioregistry (preferred) prefixes
       according to the selected strategy
    2. A python script, loaded with `PEP 723 <https://peps.python.org/pep-0723/>`_
       inline metadata, a pre-instantiated Repository object, and more
    3. A README.md file with explanation about how the code was generated, how to use
       it, etc.
    """
    if mapping_set is None and mapping_set_id is None:
        raise ValueError("either a mapping set or a mapping set ID should be given")

    import curies
    import sssom_pydantic
    from curies.vocabulary import charlie, lexical_matching_process, manual_mapping_curation
    from sssom_pydantic import MappingSet, SemanticMapping

    from ..constants import NEGATIVES_NAME, POSITIVES_NAME, PREDICTIONS_NAME, UNSURE_NAME
    from ..repository import CONFIGURATION_FILENAME, Repository

    if repository_filename is None:
        repository_filename = CONFIGURATION_FILENAME
    if positive_mappings_filename is None:
        positive_mappings_filename = POSITIVES_NAME
    if negative_mappings_filename is None:
        negative_mappings_filename = NEGATIVES_NAME
    if unsure_mappings_filename is None:
        unsure_mappings_filename = UNSURE_NAME
    if predicted_mappings_filename is None:
        predicted_mappings_filename = PREDICTIONS_NAME

    directory = Path(directory).expanduser().resolve()

    if mapping_set is None:
        mapping_set = MappingSet(
            id=mapping_set_id,
            version="1",
        )

    if mapping_set.title is None:
        mapping_set = mapping_set.model_copy(update={"title": directory.name})

    if mapping_set.license is None and add_license:
        mapping_set = mapping_set.model_copy(update={"license": CC0_CURIE})

    if not purl_base:
        purl_base, _, _ = mapping_set.id.rpartition("/")
        click.secho(
            f"`purl_base` was not given. Inferring from mapping set ID to be: {purl_base}",
            fg="yellow",
        )
        purl_base = purl_base.rstrip("/") + "/"

    converter = curies.Converter.from_prefix_map(
        {
            "ex": "https://example.org/",
            "skos": "http://www.w3.org/2004/02/skos/core#",
        }
    )
    name_to_example = {
        positive_mappings_filename: SemanticMapping(
            subject=curies.NamedReference(prefix="ex", identifier="1", name="1"),
            predicate=curies.Reference(prefix="skos", identifier="exactMatch"),
            object=curies.NamedReference(prefix="ex", identifier="2", name="2"),
            justification=manual_mapping_curation,
            authors=[charlie],
        ),
        negative_mappings_filename: SemanticMapping(
            subject=curies.NamedReference(prefix="ex", identifier="3", name="3"),
            predicate=curies.Reference(prefix="skos", identifier="exactMatch"),
            object=curies.NamedReference(prefix="ex", identifier="4", name="4"),
            justification=manual_mapping_curation,
            authors=[charlie],
        ),
        unsure_mappings_filename: SemanticMapping(
            subject=curies.NamedReference(prefix="ex", identifier="5", name="5"),
            predicate=curies.Reference(prefix="skos", identifier="exactMatch"),
            object=curies.NamedReference(prefix="ex", identifier="6", name="6"),
            justification=manual_mapping_curation,
            authors=[charlie],
        ),
        predicted_mappings_filename: SemanticMapping(
            subject=curies.NamedReference(prefix="ex", identifier="7", name="7"),
            predicate=curies.Reference(prefix="skos", identifier="exactMatch"),
            object=curies.NamedReference(prefix="ex", identifier="8", name="8"),
            justification=lexical_matching_process,
        ),
    }

    # Create the SSSOM files in a nested directory
    data_directory = directory.joinpath(DATA_DIR_NAME)
    data_directory.mkdir(exist_ok=True)
    for name, mapping in name_to_example.items():
        path = data_directory.joinpath(name)
        if path.exists():
            raise FileExistsError(f"{path} already exists. cowardly refusing to overwrite.")

        metadata = MappingSet(id=f"{purl_base}{name}")
        sssom_pydantic.write([mapping], path, metadata=metadata, converter=converter)

    data_directory_stub = Path(DATA_DIR_NAME)
    repository = Repository(
        positives_path=data_directory_stub / positive_mappings_filename,
        negatives_path=data_directory_stub / negative_mappings_filename,
        predictions_path=data_directory_stub / predicted_mappings_filename,
        unsure_path=data_directory_stub / unsure_mappings_filename,
        mapping_set=mapping_set,
        purl_base=purl_base,
    )
    repository_path = directory.joinpath(repository_filename)
    repository_path.write_text(repository.model_dump_json(indent=2, exclude=SKIPS) + "\n")

    comment = "SSSOM Curator"
    if mapping_set.title:
        comment += f" for {mapping_set.title}"

    environment = _get_jinja2_environment()
    script_template = environment.get_template("main.py.jinja2")
    script_text = script_template.render(
        comment=comment,
        repository_filename=repository_filename,
    )
    script_path = directory.joinpath(script_filename)
    script_path.write_text(script_text + "\n")

    # Get current permissions
    mode = os.stat(script_path).st_mode
    # Add user, group, and other execute bits
    os.chmod(script_path, mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)

    readme_template = environment.get_template("README.md.jinja2")
    readme_text = readme_template.render(mapping_set=mapping_set, cco_curie=CC0_CURIE)
    readme_path = directory.joinpath(readme_filename)
    readme_path.write_text(readme_text + "\n")

    if mapping_set.license == CC0_CURIE:
        license_path = directory.joinpath("LICENSE")
        license_path.write_text(HERE.joinpath("cc0.txt").read_text())

    return repository


def _get_jinja2_environment() -> jinja2.Environment:
    from jinja2 import Environment, FileSystemLoader

    environment = Environment(
        autoescape=True, loader=FileSystemLoader(HERE), trim_blocks=True, lstrip_blocks=True
    )
    return environment



"""Embedding-based mapping prediction."""

from __future__ import annotations

from collections.abc import Iterable
from typing import TYPE_CHECKING

import curies
from curies.vocabulary import lexical_matching_process
from sssom_pydantic import MappingTool, SemanticMapping
from tqdm.asyncio import tqdm

from .utils import resolve_mapping_tool

if TYPE_CHECKING:
    import pandas as pd
    from bioregistry import NormalizedNamableReference

__all__ = [
    "predict_embedding_mappings",
]


def predict_embedding_mappings(
    prefix: str,
    target_prefixes: str | Iterable[str],
    mapping_tool: str | MappingTool,
    *,
    relation: str | None | curies.NamableReference = None,
    cutoff: float | None = None,
    batch_size: int | None = None,
    progress: bool = True,
    force: bool = False,
    force_process: bool = False,
) -> list[SemanticMapping]:
    """Predict semantic mappings with embeddings."""
    import pyobo.api.embedding

    if isinstance(target_prefixes, str):
        targets = [target_prefixes]
    else:
        targets = list(target_prefixes)
    if cutoff is None:
        cutoff = 0.65
    if batch_size is None:
        batch_size = 10_000

    model = pyobo.api.embedding.get_text_embedding_model()
    source_df = pyobo.get_text_embeddings_df(
        prefix, model=model, force=force, force_process=force_process
    )

    mapping_tool = resolve_mapping_tool(mapping_tool)

    predictions = []
    for target in tqdm(targets, disable=len(targets) == 1):
        target_df = pyobo.get_text_embeddings_df(
            target, model=model, force=force, force_process=force_process
        )
        for source_id, target_id, confidence in _calculate_similarities(
            source_df, target_df, batch_size, cutoff, progress=progress
        ):
            predictions.append(
                SemanticMapping(
                    subject=_r(prefix=prefix, identifier=source_id),
                    predicate=relation,
                    object=_r(prefix=target, identifier=target_id),
                    justification=lexical_matching_process,
                    confidence=confidence,
                    mapping_tool=mapping_tool,
                )
            )
    return predictions


def _calculate_similarities(
    source_df: pd.DataFrame,
    target_df: pd.DataFrame,
    batch_size: int | None,
    cutoff: float,
    progress: bool = True,
) -> list[tuple[str, str, float]]:
    if batch_size is not None:
        return _calculate_similarities_batched(
            source_df, target_df, batch_size=batch_size, cutoff=cutoff, progress=progress
        )
    else:
        return _calculate_similarities_unbatched(source_df, target_df, cutoff=cutoff)


def _calculate_similarities_batched(
    source_df: pd.DataFrame,
    target_df: pd.DataFrame,
    *,
    batch_size: int,
    cutoff: float,
    progress: bool = True,
) -> list[tuple[str, str, float]]:
    import torch
    from sentence_transformers.util import cos_sim

    similarities = []
    source_df_numpy = source_df.to_numpy()
    for target_start in tqdm(
        range(0, len(target_df), batch_size), unit="target batch", disable=not progress
    ):
        target_end = target_start + batch_size
        target_batch_df = target_df.iloc[target_start:target_end]
        similarity = cos_sim(
            source_df_numpy,
            target_batch_df.to_numpy(),
        )
        source_target_pairs = torch.nonzero(similarity >= cutoff, as_tuple=False)
        for source_idx, target_idx in source_target_pairs:
            source_id: str = source_df.index[source_idx.item()]
            target_id: str = target_batch_df.index[target_idx.item()]
            similarities.append(
                (
                    source_id,
                    target_id,
                    similarity[source_idx, target_idx].item(),
                )
            )
    return similarities


def _calculate_similarities_unbatched(
    source_df: pd.DataFrame, target_df: pd.DataFrame, *, cutoff: float
) -> list[tuple[str, str, float]]:
    import torch
    from sentence_transformers.util import cos_sim

    similarities = []
    similarity = cos_sim(source_df.to_numpy(), target_df.to_numpy())
    source_target_pairs = torch.nonzero(similarity >= cutoff, as_tuple=False)
    for source_idx, target_idx in source_target_pairs:
        source_id: str = source_df.index[source_idx.item()]
        target_id: str = target_df.index[target_idx.item()]
        similarities.append(
            (
                source_id,
                target_id,
                similarity[source_idx, target_idx].item(),
            )
        )
    return similarities


def _r(prefix: str, identifier: str) -> NormalizedNamableReference:
    import bioregistry
    import pyobo

    return bioregistry.NormalizedNamableReference(
        prefix=prefix, identifier=identifier, name=pyobo.get_name(prefix, identifier)
    )



"""Blueprint."""

from __future__ import annotations

import getpass
from collections import Counter
from copy import deepcopy
from typing import Any, cast

import flask
import pydantic
import werkzeug
from flask import current_app
from werkzeug.local import LocalProxy

from .components import Controller, MappingForm, State
from .utils import commit, get_branch, normalize_mark, not_main, push

__all__ = [
    "blueprint",
    "url_for_state",
]


def url_for_state(endpoint: str, state: State, **kwargs: Any) -> str:
    """Get the URL for an endpoint based on the state class."""
    vv = state.model_dump(exclude_none=True, exclude_defaults=True)
    vv.update(kwargs)  # make sure stuff explicitly set overrides state
    return flask.url_for(endpoint, **vv)


CONTROLLER: Controller = cast(Controller, LocalProxy(lambda: current_app.config["controller"]))
blueprint = flask.Blueprint("ui", __name__)


@blueprint.route("/")
def home() -> str:
    """Serve the home page."""
    form = MappingForm()
    state = State.from_flask_globals()
    predictions = CONTROLLER.predictions_from_state(state)
    remaining_rows = CONTROLLER.count_predictions_from_state(state)
    return flask.render_template(
        "home.html",
        predictions=predictions,
        form=form,
        state=state,
        remaining_rows=remaining_rows,
    )


@blueprint.route("/summary")
def summary() -> str:
    """Serve the summary page."""
    state = State.from_flask_globals()
    state.limit = None
    predictions = CONTROLLER.predictions_from_state(state)
    counter = Counter((mapping.subject.prefix, mapping.object.prefix) for _, mapping in predictions)
    rows = []
    for (source_prefix, target_prefix), count in counter.most_common():
        row_state = deepcopy(state)
        row_state.source_prefix = source_prefix
        row_state.target_prefix = target_prefix
        rows.append((source_prefix, target_prefix, count, url_for_state(".home", row_state)))

    return flask.render_template(
        "summary.html",
        state=state,
        rows=rows,
    )


@blueprint.route("/add_mapping", methods=["POST"])
def add_mapping() -> werkzeug.Response:
    """Add a new mapping manually."""
    form = MappingForm()
    if form.is_submitted():
        try:
            subject = form.get_subject(CONTROLLER.converter)
        except pydantic.ValidationError as e:
            flask.flash(f"Problem with source CURIE {e}", category="warning")
            return _go_home()

        try:
            obj = form.get_object(CONTROLLER.converter)
        except pydantic.ValidationError as e:
            flask.flash(f"Problem with source CURIE {e}", category="warning")
            return _go_home()

        CONTROLLER.add_mapping(subject, obj)
        CONTROLLER.persist()
    else:
        flask.flash("missing form data", category="warning")
    return _go_home()


@blueprint.route("/commit")
def run_commit() -> werkzeug.Response:
    """Make a commit then redirect to the home page."""
    commit_info = commit(
        f"Curated {CONTROLLER.total_curated} mapping"
        f"{'s' if CONTROLLER.total_curated > 1 else ''}"
        f" ({getpass.getuser()})",
    )
    current_app.logger.warning("git commit res: %s", commit_info)
    if not_main():
        branch = get_branch()
        push_output = push(branch_name=branch)
        current_app.logger.warning("git push res: %s", push_output)
    else:
        flask.flash("did not push because on master branch")
        current_app.logger.warning("did not push because on master branch")
    CONTROLLER.total_curated = 0
    return _go_home()


@blueprint.route("/mark/<int:line>/<value>")
def mark(line: int, value: str) -> werkzeug.Response:
    """Mark the given line as correct or not."""
    CONTROLLER.mark(line, normalize_mark(value))
    CONTROLLER.persist()
    return _go_home()


def _go_home() -> werkzeug.Response:
    state = State.from_flask_globals()
    return flask.redirect(url_for_state(".home", state))



"""Command line interface for :mod:`sssom_curator`."""

import os
import sys
from pathlib import Path

import click

from .repository import CONFIGURATION_FILENAME, Repository, add_commands

__all__ = [
    "main",
]


@click.group(help="A CLI for managing SSSOM repositories.")
@click.option(
    "-p",
    "--path",
    type=click.Path(file_okay=True, dir_okay=True, exists=True),
    default=os.getcwd,
    help=f"Either the path to a sssom-curator configuration file or a directory "
    f"containing a file named {CONFIGURATION_FILENAME}. Defaults to current working directory",
)
@click.pass_context
def main(ctx: click.Context, path: Path) -> None:
    """Run the CLI."""
    if ctx.invoked_subcommand != "init":
        ctx.obj = _get_repository(path)


@main.command(name="init")
@click.option(
    "-d",
    "--directory",
    type=click.Path(file_okay=False, dir_okay=True),
    default=os.getcwd,
)
@click.option("--purl-base", help="The PURL for the exported mapping set")
@click.option("--mapping-set-title", help="The title for the mapping set")
def initialize(directory: Path, purl_base: str, mapping_set_title: str | None) -> None:
    """Initialize a repository."""
    from sssom_pydantic import MappingSet

    from .initialize import initialize_folder, normalize_name

    directory = Path(directory).resolve()
    directory.mkdir(exist_ok=True, parents=True)

    if mapping_set_title is None:
        mapping_set_title = directory.name

    click.echo(f"initialized SSSOM project `{mapping_set_title}` at `{directory.resolve()}`")

    if purl_base is None:
        norm_title = normalize_name(mapping_set_title)
        purl_base = click.prompt(
            "PURL base?", default=f"https://w3id.org/sssom/mappings/{norm_title}"
        )

    # always have the PURL end with a trailing slash
    purl_base = purl_base.rstrip("/") + "/"

    mapping_set = MappingSet(id=f"{purl_base}sssom.tsv", title=mapping_set_title, version="1")
    initialize_folder(directory, mapping_set=mapping_set, purl_base=purl_base)


def _get_repository(path: str | Path | None) -> Repository:
    if path is None:
        raise ValueError("path not given")

    path = Path(path).expanduser().resolve()
    if not path.exists():
        raise FileNotFoundError

    if path.is_file():
        return Repository.from_path(path)

    if path.is_dir():
        try:
            repository = Repository.from_directory(path)
        except FileNotFoundError as e:
            click.secho(e.args[0])
            sys.exit(1)
        else:
            return repository

    click.secho(f"bad path: {path}")
    sys.exit(1)


add_commands(main)

if __name__ == "__main__":
    main()



"""Generate a summary for the Biomappings website."""

from __future__ import annotations

import itertools as itt
import typing
from collections import Counter
from collections.abc import Iterable
from pathlib import Path
from typing import TYPE_CHECKING, Any, TypedDict

from sssom_pydantic import SemanticMapping

if TYPE_CHECKING:
    from ..repository import Repository

__all__ = [
    "summarize",
]


def summarize(
    repository: Repository,
    output: Path,
    get_orcid_to_name: typing.Callable[[], dict[str, str]] | None = None,
) -> None:
    """Do summary."""
    import yaml

    positive_mappings = repository.read_positive_mappings()
    negative_mappings = repository.read_negative_mappings()
    unsure_mappings = repository.read_unsure_mappings()
    predicted_mappings = repository.read_predicted_mappings()

    rv: dict[str, Any] = {
        "contributors": _get_contributors(
            itt.chain(positive_mappings, negative_mappings, unsure_mappings),
            get_orcid_to_name() if get_orcid_to_name is not None else {},
        ),
    }

    for key, mappings in [
        ("positive", positive_mappings),
        ("negative", negative_mappings),
        ("unsure", unsure_mappings),
        ("predictions", predicted_mappings),
    ]:
        count_records = _get_count_records(mappings)
        rv[key] = count_records
        rv[f"{key}_mapping_count"] = sum(count_record["count"] for count_record in count_records)
        rv[f"{key}_prefix_count"] = len(
            set(
                itt.chain.from_iterable(
                    (count_record["source"], count_record["target"])
                    for count_record in count_records
                )
            )
        )

    with open(output, "w") as file:
        yaml.safe_dump(rv, file, indent=2)


class CountRecord(TypedDict):
    """A result from getting counter."""

    source: str
    target: str
    count: int


def _get_count_records(mappings: Iterable[SemanticMapping]) -> list[CountRecord]:
    counter: typing.Counter[tuple[str, str]] = Counter()
    for mapping in mappings:
        subject, target = sorted([mapping.subject, mapping.object])
        counter[str(subject.prefix), str(target.prefix)] += 1
    return [
        {"source": subject_prefix, "target": target_prefix, "count": count}
        for (subject_prefix, target_prefix), count in counter.most_common()
    ]


def _get_contributors(
    mappings: Iterable[SemanticMapping], orcid_to_name: dict[str, str]
) -> list[dict[str, Any]]:
    orcid_counter = Counter(
        author.identifier
        for mapping in mappings
        if mapping.authors
        for author in mapping.authors
        if author.prefix == "orcid"
    )
    rv = []
    for orcid, count in orcid_counter.most_common():
        d = {"count": count, "orcid": orcid}
        if name := orcid_to_name.get(orcid):
            d["name"] = name
        rv.append(d)
    return rv



"""Export Biomappings as SSSOM."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import click
import curies
from sssom_pydantic import MappingSet, Metadata

if TYPE_CHECKING:
    import pandas as pd
    from sssom import MappingSetDataFrame

    from ..repository import Repository

__all__ = [
    "merge",
]

columns = [
    "subject_id",
    "subject_label",
    "predicate_id",
    "predicate_modifier",
    "object_id",
    "object_label",
    "mapping_justification",
    "author_id",
    "confidence",
    "mapping_tool",
]


def _sssom_dump(mapping_set: MappingSet) -> Metadata:
    return mapping_set.to_record().model_dump(exclude_none=True, exclude_unset=True)


def merge(repository: Repository, directory: Path) -> None:
    """Merge the SSSOM files together and output to a directory."""
    if repository.mapping_set is None:
        raise ValueError

    import yaml
    from sssom.writers import write_json, write_owl

    converter, df, msdf = get_merged_sssom(repository)

    tsv_meta = {**_sssom_dump(repository.mapping_set), "curie_map": converter.bimap}

    if repository.basename:
        fname = repository.basename
    elif repository.mapping_set.title is not None:
        fname = repository.mapping_set.title.lower().replace(" ", "-")
    else:
        raise ValueError("basename or mapping set title must be se")

    stub = directory.joinpath(fname)
    tsv_path = stub.with_suffix(".sssom.tsv")
    json_path = stub.with_suffix(".sssom.json")
    owl_path = stub.with_suffix(".sssom.owl")
    metadata_path = stub.with_suffix(".sssom.yml")

    with tsv_path.open("w") as file:
        for line in yaml.safe_dump(tsv_meta).splitlines():
            print(f"# {line}", file=file)
        df.to_csv(file, sep="\t", index=False)

    with open(metadata_path, "w") as file:
        yaml.safe_dump(tsv_meta, file)

    if not repository.purl_base:
        click.secho(
            "can not output JSON nor OWL because ``purl_base`` was not defined", fg="yellow"
        )
    else:
        _base = repository.purl_base.rstrip("/")
        click.echo("Writing JSON")
        with json_path.open("w") as file:
            msdf.metadata["mapping_set_id"] = f"{_base}/{fname}.sssom.json"
            write_json(msdf, file)
        click.echo("Writing OWL")
        with owl_path.open("w") as file:
            msdf.metadata["mapping_set_id"] = f"{_base}/{fname}.sssom.owl"
            write_owl(msdf, file)


def get_merged_sssom(
    repository: Repository, *, use_tqdm: bool = False, converter: curies.Converter | None = None
) -> tuple[curies.Converter, pd.DataFrame, MappingSetDataFrame]:
    """Get an SSSOM dataframe."""
    if repository.mapping_set is None:
        raise ValueError

    import pandas as pd
    from curies.utils import _prefix_from_curie
    from tqdm.auto import tqdm

    from ..constants import ensure_converter

    converter = ensure_converter(converter, preferred=True)
    prefixes: set[str] = {"semapv"}

    # NEW WAY: load all DFs, concat them, reorder columns

    a = pd.read_csv(repository.positives_path, sep="\t", comment="#")
    b = pd.read_csv(repository.negatives_path, sep="\t", comment="#")
    c = pd.read_csv(repository.predictions_path, sep="\t", comment="#")
    df = pd.concat([a, b, c])
    df = df[columns]

    for column in ["subject_id", "object_id", "predicate_id"]:
        converter.pd_standardize_curie(df, column=column, strict=True)

    for _, mapping in tqdm(
        df.iterrows(), desc="tabulating prefixes & authors", disable=not use_tqdm
    ):
        prefixes.add(_prefix_from_curie(mapping["subject_id"]))
        prefixes.add(_prefix_from_curie(mapping["predicate_id"]))
        prefixes.add(_prefix_from_curie(mapping["object_id"]))
        author_id = mapping["author_id"]
        if pd.notna(author_id) and any(author_id.startswith(x) for x in ["orcid:", "wikidata:"]):
            prefixes.add(_prefix_from_curie(author_id))
        # TODO add justification:

    converter = converter.get_subconverter(prefixes)

    from sssom.constants import DEFAULT_VALIDATION_TYPES
    from sssom.parsers import from_sssom_dataframe
    from sssom.validators import validate

    try:
        msdf = from_sssom_dataframe(
            df, prefix_map=converter, meta=_sssom_dump(repository.mapping_set)
        )
    except Exception as e:
        click.secho(f"SSSOM Export failed...\n{e}", fg="red")
        raise

    results = validate(msdf=msdf, validation_types=DEFAULT_VALIDATION_TYPES, fail_on_error=False)
    for validator_type, validation_report in results.items():
        if validation_report.results:
            click.secho(f"SSSOM Validator Failed: {validator_type}", fg="red")
            for result in validation_report.results:
                click.secho(f"- {result}", fg="red")
            click.echo("")

    return converter, df, msdf



"""Validation tests for :mod:`biomappings`."""

from __future__ import annotations

import unittest
from collections import defaultdict
from collections.abc import Callable, Iterable
from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar, TypeAlias, TypeVar

import curies

if TYPE_CHECKING:
    from curies import Reference
    from sssom_pydantic import SemanticMapping

    from .repository import Repository

__all__ = [
    "IntegrityTestCase",
    "MappingGetter",
    "PathIntegrityTestCase",
    "RepositoryTestCase",
]

X = TypeVar("X")
Y = TypeVar("Y")


def _extract_redundant(counter: dict[X, list[Y]]) -> list[tuple[X, list[Y]]]:
    return [(key, values) for key, values in counter.items() if len(values) > 1]


def _locations_str(locations: Iterable[tuple[Any, Any]]) -> str:
    return ", ".join(f"{label}:{line}" for label, line in locations)


#: A function that gets mappings
MappingGetter: TypeAlias = Callable[[], list["SemanticMapping"]]


class IntegrityTestCase(unittest.TestCase):
    """Data integrity tests."""

    mappings: ClassVar[list[SemanticMapping]]
    predictions: ClassVar[list[SemanticMapping]]
    incorrect: ClassVar[list[SemanticMapping]]
    unsure: ClassVar[list[SemanticMapping]]
    converter: ClassVar[curies.Converter]

    def _iter_groups(self) -> Iterable[tuple[str, int, SemanticMapping]]:
        for group, label in [
            (self.mappings, "positive"),
            (self.incorrect, "negative"),
            (self.predictions, "predictions"),
            (self.unsure, "unsure"),
        ]:
            for i, mapping in enumerate(group, start=2):
                yield label, i, mapping

    def test_mapping_justifications(self) -> None:
        """Test that the prediction type is pulled in properly."""
        from curies.vocabulary import matching_processes

        for label, line, mapping in self._iter_groups():
            self.assertEqual(
                "semapv",
                mapping.justification.prefix,
                msg=f"[{label}] The 'mapping_justification' column should "
                f"be annotated with semapv on line {line}",
            )
            self.assertIn(mapping.justification, matching_processes)

    def test_prediction_not_manual(self) -> None:
        """Test that predicted mappings don't use manual mapping justification."""
        for _line, mapping in enumerate(self.predictions, start=2):
            self.assertNotEqual(
                "ManualMappingCuration",
                mapping.justification.identifier,
                msg="Prediction can not be annotated with manual curation",
            )

    def test_valid_curies(self) -> None:
        """Test that all mappings use canonical prefixes."""
        for label, line, mapping in self._iter_groups():
            self.assert_valid(label, line, mapping.subject)
            self.assert_valid(label, line, mapping.predicate)
            self.assert_valid(label, line, mapping.object)
            self.assert_valid(label, line, mapping.justification)
            if mapping.author is not None:
                self.assert_valid(label, line, mapping.author)

    def assert_valid(self, label: str, line: int, reference: Reference) -> None:
        """Assert a reference is valid and normalized."""
        norm_prefix = self.converter.standardize_prefix(reference.prefix)
        self.assertIsNotNone(
            norm_prefix, msg=f"Unknown prefix: {reference.prefix} on {label}:{line}"
        )
        self.assertEqual(
            norm_prefix,
            reference.prefix,
            msg=f"Non-normalized prefix: {reference.prefix} on {label}:{line}",
        )
        self.assertEqual(
            self.converter.standardize_identifier(reference.prefix, reference.identifier),
            reference.identifier,
            msg=f"Invalid identifier: {reference.curie} on {label}:{line}",
        )

    def test_contributors(self) -> None:
        """Test all contributors have an entry in the curators.tsv file."""
        files = [
            ("positive", self.mappings),
            ("negative", self.incorrect),
            ("unsure", self.unsure),
        ]
        for label, mappings in files:
            for mapping in mappings:
                for author in mapping.authors or []:
                    self.assertEqual(
                        "orcid",
                        author.prefix,
                        msg=f"ORCID prefixes are required for authors in the {label} group",
                    )

    def test_cross_redundancy(self) -> None:
        """Test the redundancy of manually curated mappings and predicted mappings."""
        from sssom_pydantic.process import CanonicalMappingTuple, get_canonical_tuple

        counter: defaultdict[CanonicalMappingTuple, defaultdict[str, list[int]]] = defaultdict(
            lambda: defaultdict(list)
        )
        for label, line, mapping in self._iter_groups():
            counter[get_canonical_tuple(mapping)][label].append(line)

        redundant: list[tuple[CanonicalMappingTuple, list[tuple[str, list[int]]]]] = []
        for mapping_key, label_to_lines in counter.items():
            if len(label_to_lines) <= 1:
                continue
            redundant.append((mapping_key, sorted(label_to_lines.items())))

        if redundant:
            msg = "".join(
                f"\n  {mapping}: {_locations_str(locations)}" for mapping, locations in redundant
            )
            raise ValueError(f"{len(redundant)} are redundant: {msg}")

    def assert_no_internal_redundancies(self, mappings: list[SemanticMapping]) -> None:
        """Assert that the list of mappings doesn't have any redundancies."""
        counter: defaultdict[tuple[Reference, Reference], list[int]] = defaultdict(list)
        for line_number, mapping in enumerate(mappings, start=1):
            counter[mapping.subject, mapping.object].append(line_number)
        redundant = _extract_redundant(counter)
        if redundant:
            msg = "".join(
                f"\n  {subject.curie}/{obj.curie}: {locations}"
                for (subject, obj), locations in redundant
            )
            self.fail(f"{len(redundant)} are redundant: {msg}")

    def test_predictions_sorted(self) -> None:
        """Test the predictions are in a canonical order."""
        self.assertEqual(
            self.predictions,
            sorted(self.predictions),
            msg="Predictions are not sorted",
        )
        self.assert_no_internal_redundancies(self.predictions)

    def test_curations_sorted(self) -> None:
        """Test the true curated mappings are in a canonical order."""
        self.assertEqual(
            self.mappings,
            sorted(self.mappings),
            msg="True curations are not sorted",
        )
        self.assert_no_internal_redundancies(self.mappings)

    def test_false_mappings_sorted(self) -> None:
        """Test the false curated mappings are in a canonical order."""
        self.assertEqual(
            self.incorrect,
            sorted(self.incorrect),
            msg="False curations are not sorted",
        )
        self.assert_no_internal_redundancies(self.incorrect)

    def test_unsure_sorted(self) -> None:
        """Test the unsure mappings are in a canonical order."""
        self.assertEqual(
            self.unsure,
            sorted(self.unsure),
            msg="Unsure curations are not sorted",
        )
        self.assert_no_internal_redundancies(self.unsure)


class PathIntegrityTestCase(IntegrityTestCase):
    """A test case that can be configured with paths.

    For example, in this might be used in a custom instance of Biomappings like in the
    following:

    .. code-block:: python

        from biomappings.testing import PathIntegrityTestCase

        HERE = Path(__file__).parent.resolve()


        class TestCustom(PathIntegrityTestCase):
            predictions_path = HERE.joinpath("predictions.tsv")
            positives_path = HERE.joinpath("positive.tsv")
            negatives_path = HERE.joinpath("negative.tsv")
            unsure_path = HERE.joinpath("unsure.tsv")
    """

    predictions_path: ClassVar[str | Path]
    positives_path: ClassVar[str | Path]
    negatives_path: ClassVar[str | Path]
    unsure_path: ClassVar[str | Path]

    @classmethod
    def setUpClass(cls) -> None:
        """Set up the test case."""
        import sssom_pydantic

        cls.mappings, c1, _ = sssom_pydantic.read(cls.positives_path)
        cls.predictions, c2, _ = sssom_pydantic.read(cls.predictions_path)
        cls.incorrect, c3, _ = sssom_pydantic.read(cls.negatives_path)
        cls.unsure, c4, _ = sssom_pydantic.read(cls.unsure_path)
        cls.converter = curies.chain([c1, c2, c3, c4])


class RepositoryTestCase(IntegrityTestCase):
    """A test case that can be configured with a repository."""

    repository: ClassVar[Repository]

    @classmethod
    def setUpClass(cls) -> None:
        """Set up the test case."""
        import sssom_pydantic

        cls.mappings, c1, _ = sssom_pydantic.read(cls.repository.positives_path)
        cls.predictions, c2, _ = sssom_pydantic.read(cls.repository.predictions_path)
        cls.incorrect, c3, _ = sssom_pydantic.read(cls.repository.negatives_path)
        cls.unsure, c4, _ = sssom_pydantic.read(cls.repository.unsure_path)

        # If the converter wasn't set, then chain together the
        # converters loaded from the individual SSSOM files
        if not getattr(cls, "converter", None):
            cls.converter = curies.chain([c1, c2, c3, c4])



"""Web curation interface for :mod:`biomappings`."""

from __future__ import annotations

import os
from collections.abc import Iterable
from pathlib import Path
from typing import TYPE_CHECKING

import flask
import flask_bootstrap

from .blueprint import blueprint, url_for_state
from .components import Controller
from ..constants import DEFAULT_RESOLVER_BASE

if TYPE_CHECKING:
    from curies import Converter, Reference

__all__ = [
    "get_app",
]


def get_app(
    *,
    target_references: Iterable[Reference] | None = None,
    predictions_path: Path | None = None,
    positives_path: Path | None = None,
    negatives_path: Path | None = None,
    unsure_path: Path | None = None,
    controller: Controller | None = None,
    user: Reference | None = None,
    resolver_base: str | None = None,
    title: str | None = None,
    footer: str | None = None,
    converter: Converter | None = None,
) -> flask.Flask:
    """Get a curation flask app."""
    app = flask.Flask(__name__)
    app.config["WTF_CSRF_ENABLED"] = False
    app.config["SECRET_KEY"] = os.urandom(8)
    app.config["SHOW_RELATIONS"] = True
    app.config["SHOW_LINES"] = False
    if controller is None:
        if (
            predictions_path is None
            or positives_path is None
            or negatives_path is None
            or unsure_path is None
            or user is None
        ):
            raise ValueError
        controller = Controller(
            target_references=target_references,
            predictions_path=predictions_path,
            positives_path=positives_path,
            negatives_path=negatives_path,
            unsure_path=unsure_path,
            user=user,
            converter=converter,
        )
    if not controller._predictions and predictions_path is not None:
        raise RuntimeError(f"There are no predictions to curate in {predictions_path}")
    app.config["controller"] = controller
    flask_bootstrap.Bootstrap4(app)
    app.register_blueprint(blueprint)

    if not resolver_base:
        resolver_base = DEFAULT_RESOLVER_BASE

    app.jinja_env.globals.update(
        controller=controller,
        url_for_state=url_for_state,
        resolver_base=resolver_base,
        title=title,
        footer=footer,
    )
    return app



"""Components."""

from __future__ import annotations

from collections import defaultdict
from collections.abc import Callable, Iterable, Iterator
from pathlib import Path
from typing import Any, Literal

import curies
import flask
import sssom_pydantic
from curies import NamableReference, Reference
from curies.vocabulary import broad_match, exact_match, manual_mapping_curation, narrow_match
from flask import current_app
from flask_wtf import FlaskForm
from pydantic import BaseModel
from sssom_pydantic import SemanticMapping
from wtforms import StringField, SubmitField

from .utils import Mark
from ..constants import ensure_converter, insert

__all__ = [
    "Controller",
    "MappingForm",
    "State",
]


class State(BaseModel):
    """Contains the state for queries to the curation app."""

    limit: int | None = 10
    offset: int | None = 0
    query: str | None = None
    source_query: str | None = None
    source_prefix: str | None = None
    target_query: str | None = None
    target_prefix: str | None = None
    provenance: str | None = None
    prefix: str | None = None
    sort: str | None = None
    same_text: bool | None = None
    show_relations: bool = True
    show_lines: bool = False

    @classmethod
    def from_flask_globals(cls) -> State:
        """Get the state from the flask current request."""
        return State(
            limit=flask.request.args.get("limit", type=int, default=10),
            offset=flask.request.args.get("offset", type=int, default=0),
            query=flask.request.args.get("query"),
            source_query=flask.request.args.get("source_query"),
            source_prefix=flask.request.args.get("source_prefix"),
            target_query=flask.request.args.get("target_query"),
            target_prefix=flask.request.args.get("target_prefix"),
            provenance=flask.request.args.get("provenance"),
            prefix=flask.request.args.get("prefix"),
            sort=flask.request.args.get("sort"),
            same_text=_get_bool_arg("same_text"),
            show_relations=_get_bool_arg("show_relations") or current_app.config["SHOW_RELATIONS"],
            show_lines=_get_bool_arg("show_lines") or current_app.config["SHOW_LINES"],
        )


def _get_bool_arg(name: str) -> bool | None:
    value: str | None = flask.request.args.get(name, type=str)
    if value is not None:
        return value.lower() in {"true", "t"}
    return None


class Controller:
    """A module for interacting with the predictions and mappings."""

    _user: Reference
    _predictions: list[SemanticMapping]
    converter: curies.Converter

    def __init__(
        self,
        *,
        target_references: Iterable[Reference] | None = None,
        predictions_path: Path,
        positives_path: Path,
        negatives_path: Path,
        unsure_path: Path,
        user: Reference,
        converter: curies.Converter | None = None,
    ) -> None:
        """Instantiate the web controller.

        :param target_references: Pairs of prefix, local unique identifiers that are the
            target of curation. If this is given, pre-filters will be made before on
            predictions to only show ones where either the source or target appears in
            this set
        :param predictions_path: A custom predictions file to curate from
        :param positives_path: A custom positives file to curate to
        :param negatives_path: A custom negatives file to curate to
        :param unsure_path: A custom unsure file to curate to
        """
        self.predictions_path = predictions_path
        self._predictions, _, self._predictions_metadata = sssom_pydantic.read(
            self.predictions_path
        )

        self.positives_path = positives_path
        self.negatives_path = negatives_path
        self.unsure_path = unsure_path

        self._marked: dict[int, Mark] = {}
        self.total_curated = 0
        self._added_mappings: list[SemanticMapping] = []
        self.target_references = set(target_references or [])
        self.converter = ensure_converter(converter)

        self._current_author = user

    def _get_current_author(self) -> Reference:
        return self._current_author

    def predictions_from_state(self, state: State) -> Iterable[tuple[int, SemanticMapping]]:
        """Iterate over predictions from a state instance."""
        return self.predictions(
            offset=state.offset,
            limit=state.limit,
            query=state.query,
            source_query=state.source_query,
            source_prefix=state.source_prefix,
            target_query=state.target_query,
            target_prefix=state.target_prefix,
            prefix=state.prefix,
            sort=state.sort,
            same_text=state.same_text,
            provenance=state.provenance,
        )

    def predictions(
        self,
        *,
        offset: int | None = None,
        limit: int | None = None,
        query: str | None = None,
        source_query: str | None = None,
        source_prefix: str | None = None,
        target_query: str | None = None,
        target_prefix: str | None = None,
        prefix: str | None = None,
        sort: str | None = None,
        same_text: bool | None = None,
        provenance: str | None = None,
    ) -> Iterable[tuple[int, SemanticMapping]]:
        """Iterate over predictions.

        :param offset: If given, offset the iteration by this number
        :param limit: If given, only iterate this number of predictions.
        :param query: If given, show only equivalences that have it appearing as a
            substring in one of the source or target fields.
        :param source_query: If given, show only equivalences that have it appearing as
            a substring in one of the source fields.
        :param source_prefix: If given, show only mappings that have it appearing in the
            source prefix field
        :param target_query: If given, show only equivalences that have it appearing as
            a substring in one of the target fields.
        :param target_prefix: If given, show only mappings that have it appearing in the
            target prefix field
        :param prefix: If given, show only equivalences that have it appearing as a
            substring in one of the prefixes.
        :param same_text: If true, filter to predictions with the same label
        :param sort: If "desc", sorts in descending confidence order. If "asc", sorts in
            increasing confidence order. Otherwise, do not sort.
        :param provenance: If given, filters to provenance values matching this

        :yields: Pairs of positions and prediction dictionaries
        """
        if same_text is None:
            same_text = False
        it = self._help_it_predictions(
            query=query,
            source_query=source_query,
            source_prefix=source_prefix,
            target_query=target_query,
            target_prefix=target_prefix,
            prefix=prefix,
            sort=sort,
            same_text=same_text,
            provenance=provenance,
        )
        if offset is not None:
            try:
                for _ in range(offset):
                    next(it)
            except StopIteration:
                # if next() fails, then there are no remaining entries.
                # do not pass go, do not collect 200 euro $
                return
        if limit is None:
            yield from it
        else:
            for line_prediction, _ in zip(it, range(limit), strict=False):
                yield line_prediction

    def count_predictions_from_state(self, state: State) -> int:
        """Count the number of predictions to check for the given filters."""
        return self.count_predictions(
            query=state.query,
            source_query=state.source_query,
            source_prefix=state.source_prefix,
            target_query=state.target_query,
            target_prefix=state.target_prefix,
            prefix=state.prefix,
            same_text=state.same_text,
            provenance=state.provenance,
        )

    def count_predictions(
        self,
        query: str | None = None,
        source_query: str | None = None,
        source_prefix: str | None = None,
        target_query: str | None = None,
        target_prefix: str | None = None,
        prefix: str | None = None,
        sort: str | None = None,
        same_text: bool | None = None,
        provenance: str | None = None,
    ) -> int:
        """Count the number of predictions to check for the given filters."""
        it = self._help_it_predictions(
            query=query,
            source_query=source_query,
            source_prefix=source_prefix,
            target_query=target_query,
            target_prefix=target_prefix,
            prefix=prefix,
            sort=sort,
            same_text=same_text,
            provenance=provenance,
        )
        return sum(1 for _ in it)

    def _help_it_predictions(  # noqa:C901
        self,
        query: str | None = None,
        source_query: str | None = None,
        source_prefix: str | None = None,
        target_query: str | None = None,
        target_prefix: str | None = None,
        prefix: str | None = None,
        sort: str | None = None,
        same_text: bool | None = None,
        provenance: str | None = None,
    ) -> Iterator[tuple[int, SemanticMapping]]:
        it: Iterable[tuple[int, SemanticMapping]] = enumerate(self._predictions)
        if self.target_references:
            it = (
                (line, p)
                for (line, p) in it
                if p.subject in self.target_references or p.object in self.target_references
            )

        if query is not None:
            it = self._help_filter(
                query,
                it,
                lambda mapping: [
                    mapping.subject.curie,
                    mapping.subject_name,
                    mapping.object.curie,
                    mapping.object_name,
                    mapping.mapping_tool_name,
                ],
            )
        if source_prefix is not None:
            it = self._help_filter(source_prefix, it, lambda mapping: [mapping.subject.curie])
        if source_query is not None:
            it = self._help_filter(
                source_query, it, lambda mapping: [mapping.subject.curie, mapping.subject_name]
            )
        if target_query is not None:
            it = self._help_filter(
                target_query, it, lambda mapping: [mapping.object.curie, mapping.object_name]
            )
        if target_prefix is not None:
            it = self._help_filter(target_prefix, it, lambda mapping: [mapping.object.curie])
        if prefix is not None:
            it = self._help_filter(
                prefix, it, lambda mapping: [mapping.subject.curie, mapping.object.curie]
            )
        if provenance is not None:
            it = self._help_filter(
                provenance,
                it,
                lambda mapping: [mapping.mapping_tool_name],
            )

        def _get_confidence(t: tuple[int, SemanticMapping]) -> float:
            return t[1].confidence or 0.0

        if sort is not None:
            if sort == "desc":
                it = iter(sorted(it, key=_get_confidence, reverse=True))
            elif sort == "asc":
                it = iter(sorted(it, key=_get_confidence, reverse=False))
            elif sort == "subject":
                it = iter(sorted(it, key=lambda l_p: l_p[1].subject.curie))
            elif sort == "object":
                it = iter(sorted(it, key=lambda l_p: l_p[1].object.curie))
            else:
                raise ValueError(f"unknown sort type: {sort}")

        if same_text:
            it = (
                (line, mapping)
                for line, mapping in it
                if mapping.subject_name
                and mapping.object_name
                and mapping.subject_name.casefold() == mapping.object_name.casefold()
                and mapping.predicate.curie == "skos:exactMatch"
            )

        rv = ((line, prediction) for line, prediction in it if line not in self._marked)
        return rv

    @staticmethod
    def _help_filter(
        query: str,
        it: Iterable[tuple[int, SemanticMapping]],
        func: Callable[[SemanticMapping], list[str | None]],
    ) -> Iterable[tuple[int, SemanticMapping]]:
        query = query.casefold()
        for line, mapping in it:
            if any(query in element.casefold() for element in func(mapping) if element):
                yield line, mapping

    @property
    def total_predictions(self) -> int:
        """Return the total number of yet unmarked predictions."""
        return len(self._predictions) - len(self._marked)

    def mark(self, line: int, value: Mark) -> None:
        """Mark the given equivalency as correct.

        :param line: Position of the prediction
        :param value: Value to mark the prediction with

        :raises ValueError: if an invalid value is used
        """
        if line > len(self._predictions):
            raise IndexError(
                f"given line {line} is larger than the number of "
                f"predictions {len(self._predictions):,}"
            )
        if line not in self._marked:
            self.total_curated += 1
        self._marked[line] = value

    def add_mapping(
        self,
        subject: Reference,
        obj: Reference,
    ) -> None:
        """Add manually curated new mappings."""
        self._added_mappings.append(
            SemanticMapping.model_validate(
                {
                    "subject": subject,
                    "predicate": exact_match,
                    "object": obj,
                    "authors": [self._get_current_author()],
                    "justification": manual_mapping_curation,
                }
            )
        )
        self.total_curated += 1

    def _insert(self, mappings: Iterable[SemanticMapping], path: Path) -> None:
        insert(path=path, converter=self.converter, include_mappings=mappings)

    def persist(self) -> None:  # noqa:C901
        """Save the current markings to the source files."""
        if not self._marked:
            # no need to persist if there are no marks
            return None

        entries: defaultdict[Literal["correct", "incorrect", "unsure"], list[SemanticMapping]] = (
            defaultdict(list)
        )

        for line, value in sorted(self._marked.items(), reverse=True):
            try:
                mapping = self._predictions.pop(line)
            except IndexError:
                raise IndexError(
                    f"you tried popping the {line} element from the predictions list, "
                    f"which only has {len(self._predictions):,} elements"
                ) from None

            update: dict[str, Any] = {
                "authors": [self._get_current_author()],
                "justification": manual_mapping_curation,
                # throw the predicted confidence away, since it's been manually curated now
                "confidence": None,
            }

            entry_key: Literal["correct", "incorrect", "unsure"]
            # note these go backwards because of the way they are read
            if value == "broad":
                entry_key = "correct"
                update["predicate"] = narrow_match
            elif value == "narrow":
                entry_key = "correct"
                update["predicate"] = broad_match
            elif value == "incorrect":
                entry_key = "incorrect"
                update["predicate_modifier"] = "Not"
            elif value == "correct":
                entry_key = "correct"
            elif value == "unsure":
                entry_key = "unsure"
            else:
                raise NotImplementedError

            # replace some values using model_copy since the model is frozen
            new_mapping = mapping.model_copy(update=update)

            entries[entry_key].append(new_mapping)

        # no need to standardize since we assume everything was correct on load.
        # only write files that have some values to go in them!
        if entries["correct"]:
            self._insert(entries["correct"], path=self.positives_path)
        if entries["incorrect"]:
            self._insert(entries["incorrect"], path=self.negatives_path)
        if entries["unsure"]:
            self._insert(entries["unsure"], path=self.unsure_path)

        sssom_pydantic.write(
            self._predictions,
            self.predictions_path,
            metadata=self._predictions_metadata,
            converter=self.converter,
            drop_duplicates=True,
            sort=True,
        )
        self._marked.clear()

        # Now add manually curated mappings, if there are any
        if self._added_mappings:
            self._insert(self._added_mappings, path=self.positives_path)
            self._added_mappings = []

        return None


class MappingForm(FlaskForm):  # type:ignore[misc]
    """Form for entering new mappings."""

    # TODO how to get types for flask-wtf

    source_prefix = StringField("Source Prefix", id="source_prefix")
    source_id = StringField("Source ID", id="source_id")
    source_name = StringField("Source Name", id="source_name")
    target_prefix = StringField("Target Prefix", id="target_prefix")
    target_id = StringField("Target ID", id="target_id")
    target_name = StringField("Target Name", id="target_name")
    submit = SubmitField("Add")

    def get_subject(self, converter: curies.Converter) -> NamableReference:
        """Get the subject."""
        return NamableReference.model_validate(
            {
                "prefix": self.data["source_prefix"],
                "identifier": self.data["source_id"],
                "name": self.data["source_name"],
            },
            context=converter,
        )

    def get_object(self, converter: curies.Converter) -> NamableReference:
        """Get the object."""
        return NamableReference.model_validate(
            {
                "prefix": self.data["target_prefix"],
                "identifier": self.data["target_id"],
                "name": self.data["target_name"],
            },
            context=converter,
        )



"""Repository."""

from __future__ import annotations

import sys
import typing
from collections.abc import Callable, Iterable
from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar, Literal, TypeAlias

import click
import sssom_pydantic
from pydantic import BaseModel
from sssom_pydantic import MappingSet
from typing_extensions import Self

from .constants import (
    DEFAULT_RESOLVER_BASE,
    NEGATIVES_NAME,
    POSITIVES_NAME,
    PREDICTIONS_NAME,
    UNSURE_NAME,
    PredictionMethod,
    ensure_converter,
    insert,
)

if TYPE_CHECKING:
    import curies
    from curies import Converter
    from sssom_pydantic import MappingTool, SemanticMapping

    from .testing import IntegrityTestCase

__all__ = [
    "OrcidNameGetter",
    "Repository",
    "UserGetter",
    "add_commands",
]

#: A function that returns the current user
UserGetter: TypeAlias = Callable[[], "curies.Reference"]

#: A function that returns a dictionary from ORCID to name
OrcidNameGetter: TypeAlias = Callable[[], dict[str, str]]

#: How to decide what converter to use
ConverterStrategy: TypeAlias = Literal["bioregistry", "bioregistry-preferred", "passthrough"]

#: Configuration file
CONFIGURATION_FILENAME = "sssom-curator.json"

strategy_option = click.option(
    "--strategy",
    type=click.Choice(list(typing.get_args(ConverterStrategy))),
    default="passthrough",
    show_default=True,
)


class Repository(BaseModel):
    """A data structure containing information about a SSSOM repository.

    There are two ways to configure a repository:

    1. Parse from a JSON file representing a configuration
    2. Configure using Python

    Configuring a Repository with JSON
    ==================================

    Since the :class:`Repository` class inherits from :class:`pydantic.BaseModel`, you
    can define the data externally in a JSON file and parse it. Given the following
    example configuration (corresponding to the Biomappings project), the following
    Python code can be used to load the repository and run the CLI.

    .. code-block:: json

        {
          "predictions_path": "predictions.sssom.tsv",
          "positives_path": "positive.sssom.tsv",
          "negatives_path": "negative.sssom.tsv",
          "unsure_path": "unsure.sssom.tsv",
          "purl_base": "https://w3id.org/biopragmatics/biomappings/sssom",
          "mapping_set": {
            "mapping_set_id": "https://w3id.org/biopragmatics/biomappings/sssom/biomappings.sssom.tsv",
            "mapping_set_description": "Biomappings is a repository of community curated and predicted equivalences and related mappings between named biological entities that are not available from primary sources. It's also a place where anyone can contribute curations of predicted mappings or their own novel mappings.",
            "mapping_set_title": "Biomappings",
            "license": "https://creativecommons.org/publicdomain/zero/1.0/",
            "creator_id": ["orcid:0000-0003-4423-4370"]
          }
        }

    .. code-block:: python

        from pathlib import Path
        from sssom_curator import Repository

        path = Path("sssom-curator.json")
        repository = Repository.model_validate_json(path.read_text())

        if __name__ == "__main__":
            repository.run_cli()

    Configuring a Repository with Python
    ====================================

    You can configure your repository using the `sssom_curator.Repository` object
    directly from within Python, which offers the full flexibility of a general purpose
    programming language. Again using Biomappings as an example, here's how the Python
    file would look:

    .. code-block:: python

        from sssom_pydantic import MappingSet
        from sssom_curator import Repository
        from pathlib import Path

        # Assume files are all in the same folder
        HERE = Path(__file__).parent.resolve()

        repository = Repository(
            positives_path=HERE.joinpath("positive.sssom.tsv"),
            negatives_path=HERE.joinpath("negative.sssom.tsv"),
            unsure_path=HERE.joinpath("unsure.sssom.tsv"),
            predictions_path=HERE.joinpath("predictions.sssom.tsv"),
            mapping_set=MappingSet(
                title="Biomappings",
                id="https://w3id.org/biopragmatics/biomappings/sssom/biomappings.sssom.tsv",
            ),
            # Add the beginning part of the PURL used to
            # construct exports.
            purl_base="https://w3id.org/biopragmatics/biomappings/sssom/",
        )

        if __name__ == "__main__":
            repository.run_cli()
    """  # noqa:E501

    predictions_path: Path
    positives_path: Path
    negatives_path: Path
    unsure_path: Path
    mapping_set: MappingSet | None = None
    purl_base: str | None = None
    basename: str | None = None
    ndex_uuid: str | None = None

    web_title: str | None = None
    web_disabled_message: str | None = None
    web_footer: str | None = None

    def update_relative_paths(self, directory: Path) -> None:
        """Update paths relative to the directory."""
        if not self.predictions_path.is_file():
            self.predictions_path = directory.joinpath(self.predictions_path).resolve()
        if not self.positives_path.is_file():
            self.positives_path = directory.joinpath(self.positives_path).resolve()
        if not self.negatives_path.is_file():
            self.negatives_path = directory.joinpath(self.negatives_path).resolve()
        if not self.unsure_path.is_file():
            self.unsure_path = directory.joinpath(self.unsure_path).resolve()

    @classmethod
    def from_path(cls, path: str | Path) -> Self:
        """Load a configuration at a path."""
        path = Path(path).expanduser().resolve()
        repository = cls.model_validate_json(path.read_text())
        repository.update_relative_paths(directory=path.parent)
        return repository

    @classmethod
    def from_directory(cls, directory: str | Path) -> Self:
        """Load an implicit configuration from a directory."""
        directory = Path(directory).expanduser().resolve()
        path = directory.joinpath(CONFIGURATION_FILENAME)
        if path.is_file():
            return cls.from_path(path)

        positives_path = directory.joinpath(POSITIVES_NAME)
        negatives_path = directory.joinpath(NEGATIVES_NAME)
        predictions_path = directory.joinpath(PREDICTIONS_NAME)
        unsure_path = directory.joinpath(UNSURE_NAME)

        if (
            positives_path.is_file()
            and negatives_path.is_file()
            and predictions_path.is_file()
            and unsure_path.is_file()
        ):
            return cls(
                positives_path=positives_path,
                negatives_path=negatives_path,
                predictions_path=predictions_path,
                unsure_path=unsure_path,
            )

        raise FileNotFoundError(
            f"could not automatically construct a sssom-curator "
            f"repository from directory {directory}"
        )

    @property
    def curated_paths(self) -> list[Path]:
        """Get curated paths."""
        return [self.positives_path, self.negatives_path, self.unsure_path]

    def read_positive_mappings(self) -> list[SemanticMapping]:
        """Load the positive mappings."""
        return sssom_pydantic.read(self.positives_path)[0]

    def read_negative_mappings(self) -> list[SemanticMapping]:
        """Load the negative mappings."""
        return sssom_pydantic.read(self.negatives_path)[0]

    def read_unsure_mappings(self) -> list[SemanticMapping]:
        """Load the unsure mappings."""
        return sssom_pydantic.read(self.unsure_path)[0]

    def read_predicted_mappings(self) -> list[SemanticMapping]:
        """Load the predicted mappings."""
        return sssom_pydantic.read(self.predictions_path)[0]

    def append_positive_mappings(
        self, mappings: Iterable[SemanticMapping], *, converter: curies.Converter | None = None
    ) -> None:
        """Append new lines to the positive mappings document."""
        converter = ensure_converter(converter)
        insert(
            self.positives_path,
            converter=converter,
            include_mappings=mappings,
        )

    def append_negative_mappings(
        self, mappings: Iterable[SemanticMapping], *, converter: curies.Converter | None = None
    ) -> None:
        """Append new lines to the negative mappings document."""
        converter = ensure_converter(converter)
        insert(
            self.negatives_path,
            converter=converter,
            include_mappings=mappings,
        )

    def append_predicted_mappings(
        self, mappings: Iterable[SemanticMapping], *, converter: curies.Converter | None = None
    ) -> None:
        """Append new lines to the predicted mappings document."""
        converter = ensure_converter(converter)
        # FIXME exclude what's already in others? or is it better just
        #  to do a cleanup lint/prune step?
        insert(
            self.predictions_path,
            converter=converter,
            include_mappings=mappings,
        )

    def run_cli(self, *args: Any, **kwargs: Any) -> None:
        """Run the CLI."""
        _cli = self.get_cli()
        _cli(*args, *kwargs)

    def get_cli(
        self,
        *,
        enable_web: bool = True,
        get_user: UserGetter | None = None,
        output_directory: Path | None = None,
        sssom_directory: Path | None = None,
        image_directory: Path | None = None,
        get_orcid_to_name: OrcidNameGetter | None = None,
    ) -> click.Group:
        """Get a CLI."""

        @click.group()
        @click.pass_context
        def main(ctx: click.Context) -> None:
            """Run the CLI."""
            ctx.obj = self

        add_commands(
            main,
            enable_web=enable_web,
            get_user=get_user,
            output_directory=output_directory,
            sssom_directory=sssom_directory,
            image_directory=image_directory,
            get_orcid_to_name=get_orcid_to_name,
        )

        @main.command()
        @click.pass_context
        def update(ctx: click.Context) -> None:
            """Run all summary, merge, and chart exports."""
            click.secho("Generating summaries", fg="green")
            ctx.invoke(main.commands["summarize"])
            click.secho("Exporting SSSOM", fg="green")
            ctx.invoke(main.commands["merge"])

        return main

    def lexical_prediction_cli(
        self,
        prefix: str,
        target: str | list[str],
        /,
        *,
        mapping_tool: str | MappingTool | None = None,
        **kwargs: Any,
    ) -> None:
        """Run the lexical predictions CLI."""
        from .predict import lexical

        return lexical.lexical_prediction_cli(
            prefix,
            target,
            mapping_tool=mapping_tool,
            path=self.predictions_path,
            curated_paths=self.curated_paths,
            **kwargs,
        )

    def append_lexical_predictions(
        self,
        prefix: str,
        target_prefixes: str | Iterable[str],
        *,
        mapping_tool: str | MappingTool | None = None,
        force: bool = False,
        force_process: bool = False,
        cache: bool = True,
        converter: curies.Converter | None = None,
        **kwargs: Any,
    ) -> None:
        """Append lexical predictions."""
        from .predict import lexical

        # TODO this should reuse repository function for appending
        return lexical.append_lexical_predictions(
            prefix,
            target_prefixes,
            mapping_tool=mapping_tool,
            path=self.predictions_path,
            curated_paths=self.curated_paths,
            force=force,
            force_process=force_process,
            cache=cache,
            converter=converter,
            **kwargs,
        )

    def get_test_class(
        self,
        converter_strategy: Literal["bioregistry", "bioregistry-preferred", "passthrough"]
        | None = None,
    ) -> type[IntegrityTestCase]:
        """Get a test case class."""
        from .testing import RepositoryTestCase

        if converter_strategy is None or converter_strategy == "passthrough":

            class PassthroughTestCurator(RepositoryTestCase):
                """A test case for this repository."""

                repository: ClassVar[Repository] = self

            return PassthroughTestCurator
        elif converter_strategy == "bioregistry":

            class BioregistryTestCurator(RepositoryTestCase):
                """A test case for this repository."""

                repository: ClassVar[Repository] = self
                converter: ClassVar[Converter] = ensure_converter(preferred=False)

            return BioregistryTestCurator
        elif converter_strategy == "bioregistry-preferred":

            class BioregistryPreferredTestCurator(RepositoryTestCase):
                """A test case for this repository."""

                repository: ClassVar[Repository] = self
                converter: ClassVar[Converter] = ensure_converter(preferred=True)

            return BioregistryPreferredTestCurator
        else:
            raise ValueError(f"invalid converter strategy: {converter_strategy}")


def add_commands(
    main: click.Group,
    *,
    enable_web: bool = True,
    get_user: UserGetter | None = None,
    output_directory: Path | None = None,
    sssom_directory: Path | None = None,
    image_directory: Path | None = None,
    get_orcid_to_name: OrcidNameGetter | None = None,
) -> None:
    """Add parametrized commands."""
    main.add_command(get_lint_command())
    main.add_command(get_web_command(enable=enable_web, get_user=get_user))
    main.add_command(get_merge_command(sssom_directory=sssom_directory))
    main.add_command(get_ndex_command())
    main.add_command(
        get_summarize_command(
            output_directory=output_directory,
            image_directory=image_directory,
            get_orcid_to_name=get_orcid_to_name,
        )
    )
    main.add_command(get_predict_command())
    main.add_command(get_test_command())


def get_merge_command(sssom_directory: Path | None = None) -> click.Command:
    """Get the merge command."""

    @click.command(name="merge")
    @click.option(
        "--sssom-directory",
        type=click.Path(dir_okay=True, file_okay=False, exists=True),
        default=sssom_directory,
        required=True,
    )
    @click.pass_obj
    def main(obj: Repository, sssom_directory: Path) -> None:
        """Merge files together to a single SSSOM."""
        if sssom_directory is None:
            click.secho("--sssom-directory is required", fg="red")
            raise sys.exit(1)
        if obj.mapping_set is None:
            click.secho("repository doesn't configure ``mapping_set``", fg="red")
            raise sys.exit(1)
        if obj.purl_base is None:
            click.secho("repository doesn't configure ``purl_base``", fg="red")
            raise sys.exit(1)

        from .export.merge import merge

        merge(obj, directory=sssom_directory)

    return main


def get_summarize_command(
    output_directory: Path | None = None,
    image_directory: Path | None = None,
    get_orcid_to_name: OrcidNameGetter | None = None,
) -> click.Command:
    """Get the summary command."""

    @click.command()
    @click.option(
        "--output-directory",
        type=click.Path(file_okay=False, dir_okay=True, exists=True),
        default=output_directory.joinpath("summary.yml") if output_directory else None,
        required=True,
    )
    @click.option(
        "--image-directory",
        type=click.Path(dir_okay=True, file_okay=False),
        default=image_directory,
    )
    @click.pass_obj
    def summarize(
        obj: Repository, output_directory: Path | None, image_directory: Path | None
    ) -> None:
        """Generate summary charts and tables."""
        if output_directory is None:
            click.secho("--output-directory is required", fg="red")
            raise sys.exit(1)
        from .export.charts import make_charts
        from .export.summary import summarize

        output_directory = Path(output_directory).expanduser().resolve()
        summarize(
            obj, output_directory.joinpath("summary.yml"), get_orcid_to_name=get_orcid_to_name
        )
        make_charts(obj, output_directory, image_directory=image_directory)

    return summarize


def get_lint_command(converter: curies.Converter | None = None) -> click.Command:
    """Get the lint command."""

    @click.command()
    @strategy_option
    @click.pass_obj
    def lint(obj: Repository, strategy: ConverterStrategy) -> None:
        """Sort files and remove duplicates."""
        import sssom_pydantic

        # nonlocal lets us mess with the variable even though
        # it comes from an outside scope
        nonlocal converter
        if strategy == "passthrough":
            pass
        else:
            converter = ensure_converter(preferred=strategy == "bioregistry-preferred")

        exclude_mappings = []
        for path in obj.curated_paths:
            sssom_pydantic.lint(path, converter=converter)
            exclude_mappings.extend(sssom_pydantic.read(path)[0])

        sssom_pydantic.lint(
            obj.predictions_path,
            exclude_mappings=exclude_mappings,
            drop_duplicates=True,
        )

    return lint


def get_web_command(*, enable: bool = True, get_user: UserGetter | None = None) -> click.Command:
    """Get the web command."""
    if enable:

        @click.command()
        @click.option(
            "--resolver-base",
            help="A custom resolver base URL. Defaults to the Bioregistry.",
            default=DEFAULT_RESOLVER_BASE,
            show_default=True,
        )
        @click.option("--orcid", help="Your ORCID, if not automatically loadable")
        @click.option("--port", type=int, default=5003, show_default=True)
        @click.pass_obj
        def web(obj: Repository, resolver_base: str | None, orcid: str, port: int) -> None:
            """Run the semantic mappings curation app."""
            import webbrowser

            from curies import NamableReference
            from more_click import run_app

            from .web import get_app

            if orcid is not None:
                user = NamableReference(prefix="orcid", identifier=orcid)
            elif get_user is not None:
                user = get_user()
            else:
                orcid = (
                    click.prompt("What's your ORCID?").removeprefix("https://orcid.org").rstrip("/")
                )
                user = NamableReference(prefix="orcid", identifier=orcid)

            app = get_app(
                predictions_path=obj.predictions_path,
                positives_path=obj.positives_path,
                negatives_path=obj.negatives_path,
                unsure_path=obj.unsure_path,
                resolver_base=resolver_base,
                user=user,
                title=obj.web_title or "Semantic Mapping Curator",
                footer=obj.web_footer,
            )

            webbrowser.open_new_tab(f"http://localhost:{port}")

            run_app(app, with_gunicorn=False, port=str(port))

    else:

        @click.command()
        @click.pass_obj
        def web(obj: Repository) -> None:
            """Show an error for the web interface."""
            click.secho(
                obj.web_disabled_message
                or "web-based curator is not enabled, maybe because you're not in an editable "
                "installation of a package that build on SSSOM-Curator?",
                fg="red",
            )
            sys.exit(1)

    return web


def get_ndex_command() -> click.Command:
    """Get a CLI for uploading to NDEx."""

    @click.command()
    @click.option("--username", help="NDEx username, also looks in pystow configuration")
    @click.option("--password", help="NDEx password, also looks in pystow configuration")
    @click.pass_obj
    def ndex(obj: Repository, username: str | None, password: str | None) -> None:
        """Upload to NDEx."""
        if not obj.ndex_uuid:
            click.secho("can not upload to NDEx, no NDEx UUID is set in the curator configuration.")
            raise sys.exit(1)

        from sssom_pydantic.contrib.ndex import update_ndex

        mappings = obj.read_positive_mappings()
        update_ndex(
            uuid=obj.ndex_uuid,
            mappings=mappings,
            metadata=obj.mapping_set,
            username=username,
            password=password,
        )
        click.echo(f"Uploaded to {DEFAULT_RESOLVER_BASE}/ndex:{obj.ndex_uuid}")

    return ndex


def get_predict_command(
    *,
    source_prefix: str | None = None,
    target_prefix: str | None | list[str] = None,
) -> click.Group:
    """Create a prediction command."""
    from more_click import verbose_option

    @click.group()
    def predict() -> None:
        """Predict semantic mappings."""

    if source_prefix is None:
        source_prefix_argument = click.argument("source_prefix")
    else:
        source_prefix_argument = click.option("--source-prefix", default=source_prefix)

    if target_prefix is None:
        target_prefix_argument = click.argument("target_prefix", nargs=-1)
    else:
        target_prefix_argument = click.option(
            "--target-prefix", multiple=True, default=[target_prefix]
        )

    @predict.command()
    @verbose_option
    @source_prefix_argument
    @target_prefix_argument
    @click.option("--relation", help="the predicate to assign to semantic mappings")
    @click.option(
        "--method",
        type=click.Choice(list(typing.get_args(PredictionMethod))),
        help="The prediction method to use",
    )
    @click.option(
        "--cutoff",
        type=float,
        help="The cosine similarity cutoff to use for calling mappings when "
        "using embedding predictions",
    )
    @click.option(
        "--filter-mutual-mappings",
        is_flag=True,
        help="Remove predictions that correspond to already existing mappings "
        "in either the subject or object resource",
    )
    @click.option(
        "--force", is_flag=True, help="Force re-downloading and re-processing of resources"
    )
    @click.option(
        "--force-process",
        is_flag=True,
        help="Force re-processing, but not re-downloading of resources",
    )
    @click.option(
        "--cache/--no-cache",
        is_flag=True,
        help="Should a cache be made",
    )
    @click.option(
        "--all-by-all",
        is_flag=True,
        help="Don't just predict from source to targets, but also between all targets",
    )
    @click.pass_obj
    def lexical(
        obj: Repository,
        source_prefix: str,
        target_prefix: str,
        relation: str | None,
        method: PredictionMethod | None,
        cutoff: float | None,
        filter_mutual_mappings: bool,
        cache: bool,
        force: bool,
        force_process: bool,
        all_by_all: bool,
    ) -> None:
        """Predict semantic mappings with lexical methods."""
        from .predict.lexical import append_lexical_predictions

        append_lexical_predictions(
            source_prefix,
            target_prefix,
            path=obj.predictions_path,
            curated_paths=obj.curated_paths,
            filter_mutual_mappings=filter_mutual_mappings,
            relation=relation,
            method=method,
            cutoff=cutoff,
            cache=cache,
            force=force,
            force_process=force_process,
            all_by_all=all_by_all,
        )

    return predict


def get_test_command() -> click.Command:
    """Get a command to run tests."""

    @click.command()
    @strategy_option
    @click.pass_obj
    def test(obj: Repository, strategy: ConverterStrategy) -> None:
        """Test the repository."""
        import unittest

        test_case_class = obj.get_test_class(converter_strategy=strategy)
        loader = unittest.TestLoader()
        suite = loader.loadTestsFromTestCase(test_case_class)

        runner = unittest.TextTestRunner(verbosity=2)
        result = runner.run(suite)

        # Exit with code 1 if tests failed, 0 otherwise
        sys.exit(not result.wasSuccessful())

    return test



"""Utilities for the web app."""

from __future__ import annotations

from typing import Literal, TypeAlias, TypeVar, get_args

__all__ = [
    "Mark",
    "commit",
    "get_branch",
    "normalize_mark",
    "not_main",
    "push",
]

X = TypeVar("X")
Y = TypeVar("Y")


def commit(message: str) -> str | None:
    """Make a commit with the following message."""
    return _git("commit", "-m", message, "-a")


def push(branch_name: str | None = None) -> str | None:
    """Push the git repo."""
    if branch_name is not None:
        return _git("push", "origin", branch_name)
    else:
        return _git("push")


def not_main() -> bool:
    """Return if on the master branch."""
    return "master" != _git("rev-parse", "--abbrev-ref", "HEAD")


def get_branch() -> str:
    """Return current git branch."""
    rv = _git("branch", "--show-current")
    if rv is None:
        raise RuntimeError
    return rv


def _git(*args: str) -> str | None:
    import os
    from subprocess import CalledProcessError, check_output

    with open(os.devnull, "w") as devnull:
        try:
            ret = check_output(  # noqa: S603
                ["git", *args],  # noqa:S607
                cwd=os.path.dirname(__file__),
                stderr=devnull,
            )
        except CalledProcessError as e:
            print(e)  # noqa:T201
            return None
        else:
            return ret.strip().decode("utf-8")


Mark: TypeAlias = Literal["correct", "incorrect", "unsure", "broad", "narrow"]
MARKS: set[Mark] = set(get_args(Mark))
CORRECT = {"yup", "true", "t", "correct", "right", "close enough", "disco"}
INCORRECT = {"no", "nope", "false", "f", "nada", "nein", "incorrect", "negative", "negatory"}
UNSURE = {"unsure", "maybe", "idk", "idgaf", "idgaff"}


def normalize_mark(value: str) -> Mark:
    """Get the mark."""
    value = value.lower()
    if value in CORRECT:
        return "correct"
    elif value in INCORRECT:
        return "incorrect"
    elif value in UNSURE:
        return "unsure"
    elif value in {"broader", "broad"}:
        return "broad"
    elif value in {"narrow", "narrower"}:
        return "narrow"
    else:
        raise ValueError



"""Constants for sssom-curator."""

from __future__ import annotations

from collections.abc import Iterable
from functools import lru_cache
from pathlib import Path
from typing import TYPE_CHECKING, Literal, TypeAlias

if TYPE_CHECKING:
    import curies
    from sssom_pydantic import SemanticMapping

__all__ = [
    "DEFAULT_RESOLVER_BASE",
    "NEGATIVES_NAME",
    "POSITIVES_NAME",
    "PREDICTIONS_NAME",
    "UNSURE_NAME",
    "PredictionMethod",
    "RecognitionMethod",
    "ensure_converter",
    "insert",
]


RecognitionMethod: TypeAlias = Literal["ner", "grounding"]
PredictionMethod: TypeAlias = Literal["ner", "grounding", "embedding"]

DEFAULT_RESOLVER_BASE = "https://bioregistry.io"


def ensure_converter(
    converter: curies.Converter | None = None, *, preferred: bool = False
) -> curies.Converter:
    """Get a converter."""
    if converter is not None:
        return converter
    try:
        import bioregistry
    except ImportError as e:
        raise ImportError(
            "No converter was given, and could not import the Bioregistry. "
            "Install with:\n\n\t$ pip install bioregistry"
        ) from e

    if preferred:
        return _get_preferred()
    else:
        return bioregistry.get_default_converter()


@lru_cache(1)
def _get_preferred() -> curies.Converter:
    import bioregistry

    return bioregistry.get_converter(
        uri_prefix_priority=["rdf", "default"],
        prefix_priority=["preferred", "default"],
    )


PREDICTIONS_NAME = "predictions.sssom.tsv"
POSITIVES_NAME = "positive.sssom.tsv"
NEGATIVES_NAME = "negative.sssom.tsv"
UNSURE_NAME = "unsure.sssom.tsv"

STUB_SSSOM_COLUMNS = [
    "subject_id",
    "subject_label",
    "predicate_id",
    "object_id",
    "object_label",
    "mapping_justification",
    "author_id",
    "mapping_tool",
    "predicate_modifier",
]


def insert(
    path: Path,
    *,
    converter: curies.Converter | None = None,
    include_mappings: Iterable[SemanticMapping] | None = None,
) -> None:
    """Append eagerly with linting at the same time."""
    import sssom_pydantic

    mappings, converter_processed, metadata = sssom_pydantic.read(path, converter=converter)

    if include_mappings is not None:
        prefixes: set[str] = set()
        for mapping in include_mappings:
            prefixes.update(mapping.get_prefixes())
            mappings.append(mapping)

        for prefix in prefixes:
            if not converter_processed.standardize_prefix(prefix):
                raise NotImplementedError("amending prefixes not yet implemented")

    sssom_pydantic.write(
        mappings,
        path,
        converter=converter_processed,
        metadata=metadata,
        sort=True,
        drop_duplicates=True,
    )
